# -*- coding: utf-8 -*-
"""
Laws Panel
==========
Bibliometric laws analysis: Lotka, Bradford, Zipf, Price, Pareto.
"""

import tkinter as tk
from tkinter import ttk, messagebox
import threading
from typing import Dict, Optional

from biblium.gui.config import FONTS, LAYOUT, get_theme
from biblium.gui.core.events import event_bus, EventBus
from biblium.gui.panels.base import BasePanel
from biblium.gui.widgets.cards import Card, StatsCard, CardGrid
from biblium.gui.widgets.buttons import ThemedButton, ActionButton
from biblium.gui.widgets.forms import LabeledCombobox, LabeledCheckbox, LabeledSpinbox
from biblium.gui.widgets.tables import DataTable
from biblium.gui.widgets.plots import PlotFrame

try:
    import pandas as pd
    import numpy as np
    HAS_NUMPY = True
except ImportError:
    HAS_NUMPY = False

try:
    from scipy import stats as scipy_stats
    from scipy.optimize import curve_fit
    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False

try:
    import matplotlib.pyplot as plt
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


class LawsPanel(BasePanel):
    """Panel for bibliometric laws analysis."""
    
    title = "Bibliometric Laws"
    icon = "📚"
    description = "Analyze Lotka, Bradford, Zipf, and other bibliometric laws"
    requires_data = True
    
    def __init__(self, parent, theme: str = "light", bib=None, **kwargs):
        super().__init__(parent, theme=theme, bib=bib, **kwargs)
        self._primary_action = self._analyze_law  # Set primary action for toolbar Run button
    
    def _create_options(self):
        """Create the options panel."""
        self._add_title()
        
        # Law Selection
        law_card = Card(self.options_content, title="📚 Select Law", theme=self.theme_name)
        law_card.pack(fill=tk.X, padx=8, pady=8)
        
        self.law_combo = LabeledCombobox(
            law_card.content, label="Law:",
            values=[
                "Lotka's Law (Author Productivity)",
                "Bradford's Law (Journal Scatter)",
                "Zipf's Law (Word Frequency)",
                "Price's Law (Elite Productivity)",
                "Pareto Principle (80/20 Rule)",
            ],
            default="Lotka's Law (Author Productivity)",
            theme=self.theme_name, label_width=12,
        )
        self.law_combo.pack(fill=tk.X, pady=4)
        
        # Options
        options_card = Card(self.options_content, title="⚙️ Options", theme=self.theme_name)
        options_card.pack(fill=tk.X, padx=8, pady=8)
        
        self.entity_combo = LabeledCombobox(
            options_card.content, label="Entity:",
            values=["Authors", "Sources", "Keywords"],
            default="Authors", theme=self.theme_name, label_width=12,
        )
        self.entity_combo.pack(fill=tk.X, pady=4)
        
        self.show_fit_cb = LabeledCheckbox(
            options_card.content, label="Show theoretical fit",
            default=True, theme=self.theme_name,
        )
        self.show_fit_cb.pack(fill=tk.X, pady=2)
        
        self.log_scale_cb = LabeledCheckbox(
            options_card.content, label="Use log-log scale",
            default=True, theme=self.theme_name,
        )
        self.log_scale_cb.pack(fill=tk.X, pady=2)
        
        # Run Button
        # LLM Configuration (v2.11)
        self._create_llm_config_card()
        
        btn_frame = tk.Frame(self.options_content, bg=self.theme["bg_secondary"])
        btn_frame.pack(fill=tk.X, padx=8, pady=16)
        
        ActionButton(
            btn_frame, text="Analyze Law", icon="📊",
            command=self._analyze_law, theme=self.theme_name,
        ).pack(fill=tk.X)
    
    def _create_results(self):
        """Create results panel."""
        super()._create_results()
        
        # Create notebook with Info tab
        self.results_notebook = ttk.Notebook(self.results_content)
        self.results_notebook.pack(fill=tk.BOTH, expand=True)
        
        # Results tab (will be populated by analysis)
        self.results_tab = tk.Frame(self.results_notebook, bg=self.theme["bg_card"])
        self.results_notebook.add(self.results_tab, text="📊 Results")
        
        # Info tab
        self.info_tab = tk.Frame(self.results_notebook, bg=self.theme["bg_card"])
        self.results_notebook.add(self.info_tab, text="ℹ️ Info")
        self._create_info_content(self.info_tab)
        
        # Show placeholder in results tab
        self._show_placeholder_in_tab()
    
    def _show_placeholder_in_tab(self):
        """Show placeholder in results tab."""
        # Check if results_tab still exists
        if not hasattr(self, 'results_tab'):
            return
        try:
            if not self.results_tab.winfo_exists():
                return
        except tk.TclError:
            return
            
        for widget in self.results_tab.winfo_children():
            widget.destroy()
        tk.Label(
            self.results_tab,
            text="📚 Bibliometric Laws\n\nSelect a law and click 'Analyze Law'\nto test bibliometric regularities.",
            font=FONTS.get_font("body"),
            bg=self.theme["bg_card"],
            fg=self.theme["text_muted"],
            justify=tk.CENTER,
        ).pack(expand=True)
    
    def _show_loading(self, message: str = "Analyzing..."):
        """Show loading state inside the results tab, preserving notebook structure."""
        self._stop_active_spinners()
        
        # Only clear the results_tab content, not the whole notebook
        if self._widget_exists(self.results_tab):
            try:
                for widget in self.results_tab.winfo_children():
                    try:
                        widget.destroy()
                    except:
                        pass
            except tk.TclError:
                pass
            
            # Create loading indicator inside results_tab
            try:
                from biblium.gui.widgets.progress import LoadingSpinner
                
                loading_frame = tk.Frame(self.results_tab, bg=self.theme["bg_card"])
                loading_frame.pack(expand=True)
                
                spinner = LoadingSpinner(loading_frame, size=32, theme=self.theme_name)
                spinner.pack()
                spinner.start()
                
                # Track active spinner for cleanup
                if not hasattr(self, '_active_spinners'):
                    self._active_spinners = []
                self._active_spinners.append(spinner)
                
                tk.Label(
                    loading_frame,
                    text=message,
                    font=FONTS.get_font("body"),
                    bg=self.theme["bg_card"],
                    fg=self.theme["text_muted"],
                ).pack(pady=(12, 0))
                
                # Switch to results tab to show loading
                self.results_notebook.select(0)
            except tk.TclError:
                pass
        else:
            # Fallback to base behavior if notebook structure is missing
            super()._show_loading(message)
    
    def _create_info_content(self, parent):
        """Create Info tab with documentation."""
        info_text = """
BIBLIOMETRIC LAWS
═════════════════

Test classical bibliometric regularities on your data.

LOTKA'S LAW (1926)
──────────────────
Author productivity distribution:
"The number of authors making n contributions 
is about 1/n² of those making one."

Formula: y = C / x^α
• x = number of publications
• y = number of authors
• α ≈ 2 (Lotka's exponent)
• C = constant

Output:
• Observed vs expected frequencies
• Alpha exponent estimation
• Kolmogorov-Smirnov goodness-of-fit
• Chi-square test

BRADFORD'S LAW (1934)
─────────────────────
Journal scatter:
"Journals can be divided into zones of equal 
yield with counts 1 : n : n²"

• Zone 1 (Core): Few journals, many papers
• Zone 2: More journals, same papers
• Zone 3 (Periphery): Many journals, few papers

Output:
• Zone boundaries
• Core journal list
• Bradford multiplier (n)
• Cumulative distribution plot

ZIPF'S LAW (1935)
─────────────────
Word frequency:
"Frequency × rank ≈ constant"

Formula: f(r) = C / r^α
• r = word rank
• f = frequency
• α ≈ 1 (Zipf's exponent)

Output:
• Rank-frequency distribution
• Alpha estimation
• Log-log plot
• Goodness-of-fit

PRICE'S LAW
───────────
"Half of publications come from √N authors"
• N = total number of authors
• Top √N authors = 50% of output

PARETO PRINCIPLE
────────────────
"80% of effects from 20% of causes"
• 20% of authors → 80% of papers
• 20% of journals → 80% of articles
"""
        text_widget = tk.Text(
            parent, font=FONTS.get_font("body"),
            bg=self.theme["bg_card"], fg=self.theme["text_primary"],
            wrap=tk.WORD, padx=16, pady=16, relief=tk.FLAT,
        )
        text_widget.insert("1.0", info_text)
        text_widget.config(state=tk.DISABLED)
        text_widget.pack(fill=tk.BOTH, expand=True)
        
        # Add right-click context menu for copy
        def show_context_menu(event):
            menu = tk.Menu(text_widget, tearoff=0)
            menu.add_command(label="Copy Selected", command=lambda: copy_selected(text_widget))
            menu.add_command(label="Copy All", command=lambda: copy_all(text_widget))
            menu.tk_popup(event.x_root, event.y_root)
        
        def copy_selected(widget):
            try:
                widget.config(state=tk.NORMAL)
                selected = widget.get(tk.SEL_FIRST, tk.SEL_LAST)
                widget.config(state=tk.DISABLED)
                widget.clipboard_clear()
                widget.clipboard_append(selected)
            except tk.TclError:
                pass
        
        def copy_all(widget):
            widget.config(state=tk.NORMAL)
            content = widget.get("1.0", tk.END)
            widget.config(state=tk.DISABLED)
            widget.clipboard_clear()
            widget.clipboard_append(content.strip())
        
        text_widget.bind("<Button-3>", show_context_menu)
        text_widget.bind("<Control-c>", lambda e: copy_selected(text_widget))
    
    def _analyze_law(self):
        """Run law analysis using biblium methods."""
        if not self.bib:
            messagebox.showwarning("No Data", "Please load a dataset first.")
            return
        
        self._show_loading("Analyzing...")
        
        law = self.law_combo.get()
        entity = self.entity_combo.get()
        
        def do_analysis():
            try:
                result = {"law": law, "entity": entity}
                
                # Helper to find column in dataframe
                def find_column(candidates):
                    """Find first existing column from candidates list."""
                    for col in candidates:
                        if col in self.bib.df.columns:
                            return col
                    return None
                
                # Import utilsbib for direct computation
                from biblium import utilsbib
                
                if "Lotka" in law:
                    # Find author column - WoS uses "Authors or Inventors"
                    author_col = find_column([
                        "Authors", "Authors or Inventors", "Author full names",
                        "AU", "author", "authors"
                    ])
                    
                    if author_col is None:
                        raise ValueError("No author column found in dataset")
                    
                    # Compute Lotka distribution directly (no plotting)
                    lotka_df = utilsbib.compute_lotka_distribution(
                        self.bib.df,
                        author_col=author_col,
                        separator=self.bib.default_separator,
                    )
                    lotka_stats = utilsbib.evaluate_lotka_fit(lotka_df)
                    
                    # Store in bib object for consistency
                    self.bib.lotka_df = lotka_df
                    self.bib.lotka_stats_df = lotka_stats
                    
                    if lotka_df is not None and len(lotka_df) > 0:
                        # Rename columns for display
                        result["data"] = lotka_df.rename(columns={
                            'n_pubs': 'Publications',
                            'n_authors': 'Authors',
                            'expected_n_authors': 'Expected'
                        })
                        
                        # Build interpretation from stats
                        interpretation = "Lotka's Law: Authors with n publications ≈ C/n²"
                        if lotka_stats is not None and len(lotka_stats) > 0:
                            stats_dict = dict(zip(lotka_stats['Measure'], lotka_stats['Value']))
                            r2 = stats_dict.get('R2', None)
                            if r2 is not None:
                                interpretation += f"\nR² = {r2:.3f}"
                            ks_p = stats_dict.get('KS_pvalue', None)
                            if ks_p is not None:
                                interpretation += f", KS p-value = {ks_p:.4f}"
                        
                        result["interpretation"] = interpretation
                        result["stats_df"] = lotka_stats
                        result["from_biblium"] = True
                    else:
                        result = self._fallback_lotka(entity)
                
                elif "Bradford" in law:
                    # Find source column
                    source_col = find_column([
                        "Source title", "Source Title", "Journal", "Publication Name",
                        "SO", "source", "journal"
                    ])
                    
                    if source_col is None:
                        raise ValueError("No source/journal column found in dataset")
                    
                    # Compute Bradford distribution directly (no plotting)
                    bradford_df = utilsbib.compute_bradford_distribution(
                        self.bib.df,
                        source_col=source_col,
                    )
                    bradford_stats = utilsbib.evaluate_bradford_fit(bradford_df)
                    
                    # Store in bib object for consistency
                    self.bib.bradford_df = bradford_df
                    self.bib.bradford_stats_df = bradford_stats
                    
                    if bradford_df is not None and len(bradford_df) > 0:
                        # Add rank column and rename for display
                        display_df = bradford_df.copy()
                        display_df.insert(0, 'Rank', range(1, len(display_df) + 1))
                        display_df = display_df.rename(columns={
                            'Document_Count': 'Articles',
                            'Cumulative_Documents': 'Cumulative',
                            'Cumulative_Percentage': 'Cum %'
                        })
                        result["data"] = display_df.head(100)
                        
                        interpretation = "Bradford's Law divides sources into zones with equal articles but increasing journals."
                        if bradford_stats is not None and len(bradford_stats) > 0:
                            mean_dev = bradford_stats['Mean Deviation'].iloc[0] if 'Mean Deviation' in bradford_stats.columns else None
                            if mean_dev is not None:
                                interpretation += f"\nMean deviation from expected: {mean_dev:.3f}"
                        
                        result["interpretation"] = interpretation
                        result["stats_df"] = bradford_stats
                        result["from_biblium"] = True
                    else:
                        result = self._fallback_bradford(entity)
                    
                elif "Zipf" in law:
                    # Use biblium's zipf_law method or fallback
                    # First check if abstract column exists
                    abstract_col = find_column([
                        "Processed Abstract", "Abstract", "AB", "abstract",
                        "Description", "Summary"
                    ])
                    
                    # For keywords, use keywords instead
                    if entity == "Keywords":
                        kw_col = find_column([
                            "Author Keywords", "Keywords", "DE", "ID",
                            "Index Keywords", "Processed Author Keywords"
                        ])
                        if kw_col:
                            # Use fallback for keywords - simpler approach
                            result = self._fallback_zipf_keywords()
                        else:
                            raise ValueError("No keyword column found in dataset")
                    elif abstract_col:
                        # Try using biblium's zipf_law
                        try:
                            self.bib.zipf_law(items='words from abstract', show=False)
                            zipf_df = getattr(self.bib, 'zipf_df', None)
                            zipf_stats = getattr(self.bib, 'zipf_stats', None)
                            
                            if zipf_df is not None and len(zipf_df) > 0:
                                display_df = zipf_df.head(100).copy()
                                display_df = display_df.rename(columns={
                                    'Word': 'Item',
                                    'Frequency': 'Frequency',
                                    'Rank': 'Rank'
                                })
                                if 'Rank' in display_df.columns:
                                    cols = ['Rank'] + [c for c in display_df.columns if c != 'Rank']
                                    display_df = display_df[cols]
                                
                                result["data"] = display_df
                                interpretation = "Zipf's Law: Word frequency is inversely proportional to rank."
                                if zipf_stats is not None:
                                    r2 = zipf_stats.get('R2', None)
                                    if r2 is not None:
                                        interpretation += f"\nR² = {r2:.3f}"
                                result["interpretation"] = interpretation
                                result["from_biblium"] = True
                            else:
                                result = self._fallback_zipf_abstract(abstract_col)
                        except Exception:
                            result = self._fallback_zipf_abstract(abstract_col)
                    else:
                        # No abstract - try title
                        title_col = find_column(["Title", "TI", "title", "Article Title"])
                        if title_col:
                            result = self._fallback_zipf_title(title_col)
                        else:
                            raise ValueError("No text column (Abstract/Title) found for Zipf analysis")
                    
                elif "Price" in law:
                    # Price's Law - use biblium counts if available
                    result = self._analyze_price_law(entity)
                    
                elif "Pareto" in law:
                    # Pareto Principle - use biblium counts if available
                    result = self._analyze_pareto(entity)
                
                self.after(0, lambda r=result: self._on_analysis_success(r))
            except Exception as e:
                import traceback
                traceback.print_exc()
                self.after(0, lambda msg=str(e): self._on_analysis_error(msg))
        
        threading.Thread(target=do_analysis, daemon=True).start()
    
    def _fallback_lotka(self, entity):
        """Fallback for Lotka's law when biblium method fails."""
        df = self.bib.df
        
        # Find appropriate column
        if entity == "Authors":
            col = None
            for c in ["Authors", "Authors or Inventors", "Author full names", "AU"]:
                if c in df.columns:
                    col = c
                    break
        elif entity == "Sources":
            col = None
            for c in ["Source title", "Source Title", "Journal", "Publication Name", "SO"]:
                if c in df.columns:
                    col = c
                    break
        else:
            col = None
            for c in ["Author Keywords", "Keywords", "DE", "ID"]:
                if c in df.columns:
                    col = c
                    break
        
        if col is None:
            raise ValueError(f"No {entity} column found")
        
        if entity in ["Authors", "Keywords"]:
            all_items = df[col].dropna().str.split(";").explode().str.strip()
        else:
            all_items = df[col].dropna()
        
        item_counts = all_items.value_counts()
        productivity = item_counts.value_counts().sort_index()
        x = productivity.index.values.astype(float)
        y = productivity.values.astype(float)
        
        c = y[0] if len(y) > 0 else 1
        n = 2.0
        expected = c / (x ** n)
        
        return {
            "law": "Lotka's Law (Author Productivity)",
            "entity": entity,
            "data": pd.DataFrame({
                "Publications": x.astype(int),
                "Authors": y.astype(int),
                "Expected": expected.round(1),
            }),
            "interpretation": f"Lotka's Law: Authors with n publications ≈ C/n². Exponent n ≈ {n:.2f}"
        }
    
    def _fallback_bradford(self, entity):
        """Fallback for Bradford's law when biblium method fails."""
        df = self.bib.df
        
        # Find source column
        col = None
        for c in ["Source title", "Source Title", "Journal", "Publication Name", "SO"]:
            if c in df.columns:
                col = c
                break
        
        if col is None:
            raise ValueError("No source/journal column found")
        
        item_counts = df[col].dropna().value_counts()
        sorted_counts = item_counts.sort_values(ascending=False)
        cumsum = sorted_counts.cumsum()
        
        return {
            "law": "Bradford's Law (Journal Scatter)",
            "entity": entity,
            "data": pd.DataFrame({
                "Rank": range(1, min(101, len(sorted_counts) + 1)),
                "Source": sorted_counts.index[:100],
                "Articles": sorted_counts.values[:100],
                "Cumulative": cumsum.values[:100],
            }),
            "interpretation": "Bradford's Law divides sources into zones with equal articles but increasing journals."
        }
    
    def _fallback_zipf(self, entity):
        """Fallback for Zipf's law when biblium method fails."""
        df = self.bib.df
        
        if entity == "Keywords":
            col = self.bib.mapping.get("Author_Keywords", "Author Keywords")
        else:
            col = self.bib.mapping.get("Authors", "Authors")
        
        if col not in df.columns:
            raise ValueError(f"Column {col} not found")
        
        if entity in ["Authors", "Keywords"]:
            all_items = df[col].dropna().str.split(";").explode().str.strip()
        else:
            all_items = df[col].dropna()
        
        item_counts = all_items.value_counts()
        sorted_counts = item_counts.sort_values(ascending=False)
        ranks = np.arange(1, len(sorted_counts) + 1)
        freqs = sorted_counts.values
        
        c = freqs[0] if len(freqs) > 0 else 1
        expected = c / ranks
        
        return {
            "law": "Zipf's Law (Word Frequency)",
            "entity": entity,
            "data": pd.DataFrame({
                "Rank": ranks[:100],
                "Item": sorted_counts.index[:100],
                "Frequency": freqs[:100],
                "Expected": expected[:100].round(1),
            }),
            "interpretation": "Zipf's Law: Word frequency is inversely proportional to rank."
        }
    
    def _fallback_zipf_keywords(self):
        """Zipf analysis using keywords."""
        df = self.bib.df
        
        # Find keyword column
        kw_col = None
        for col in ["Author Keywords", "Keywords", "DE", "ID", "Index Keywords"]:
            if col in df.columns:
                kw_col = col
                break
        
        if kw_col is None:
            raise ValueError("No keyword column found")
        
        # Split and count keywords
        all_kws = df[kw_col].dropna().str.split(";").explode().str.strip().str.lower()
        all_kws = all_kws[all_kws != '']
        kw_counts = all_kws.value_counts()
        
        ranks = np.arange(1, len(kw_counts) + 1)
        freqs = kw_counts.values
        c = freqs[0] if len(freqs) > 0 else 1
        expected = c / ranks
        
        return {
            "law": "Zipf's Law (Word Frequency)",
            "entity": "Keywords",
            "data": pd.DataFrame({
                "Rank": ranks[:100],
                "Item": kw_counts.index[:100],
                "Frequency": freqs[:100],
                "Expected": expected[:100].round(1),
            }),
            "interpretation": "Zipf's Law: Keyword frequency is inversely proportional to rank."
        }
    
    def _fallback_zipf_abstract(self, abstract_col):
        """Zipf analysis using abstract text."""
        import re
        df = self.bib.df
        
        # Extract words from abstracts
        all_text = df[abstract_col].dropna().str.cat(sep=' ')
        # Simple tokenization
        words = re.findall(r'\b[a-zA-Z]{3,}\b', all_text.lower())
        
        # Remove common stopwords
        stopwords = {'the', 'and', 'for', 'with', 'that', 'this', 'from', 'are', 'was', 
                     'were', 'been', 'have', 'has', 'had', 'not', 'but', 'can', 'which',
                     'their', 'they', 'these', 'there', 'than', 'also', 'into', 'such',
                     'more', 'other', 'some', 'what', 'when', 'where', 'who', 'how'}
        words = [w for w in words if w not in stopwords]
        
        word_counts = pd.Series(words).value_counts()
        
        ranks = np.arange(1, len(word_counts) + 1)
        freqs = word_counts.values
        c = freqs[0] if len(freqs) > 0 else 1
        expected = c / ranks
        
        return {
            "law": "Zipf's Law (Word Frequency)",
            "entity": "Words from Abstract",
            "data": pd.DataFrame({
                "Rank": ranks[:100],
                "Item": word_counts.index[:100],
                "Frequency": freqs[:100],
                "Expected": expected[:100].round(1),
            }),
            "interpretation": "Zipf's Law: Word frequency from abstracts is inversely proportional to rank."
        }
    
    def _fallback_zipf_title(self, title_col):
        """Zipf analysis using title text."""
        import re
        df = self.bib.df
        
        # Extract words from titles
        all_text = df[title_col].dropna().str.cat(sep=' ')
        words = re.findall(r'\b[a-zA-Z]{3,}\b', all_text.lower())
        
        # Remove common stopwords
        stopwords = {'the', 'and', 'for', 'with', 'that', 'this', 'from', 'are', 'was', 
                     'were', 'been', 'have', 'has', 'had', 'not', 'but', 'can', 'which',
                     'their', 'they', 'these', 'there', 'than', 'also', 'into', 'such',
                     'more', 'other', 'some', 'what', 'when', 'where', 'who', 'how'}
        words = [w for w in words if w not in stopwords]
        
        word_counts = pd.Series(words).value_counts()
        
        ranks = np.arange(1, len(word_counts) + 1)
        freqs = word_counts.values
        c = freqs[0] if len(freqs) > 0 else 1
        expected = c / ranks
        
        return {
            "law": "Zipf's Law (Word Frequency)",
            "entity": "Words from Title",
            "data": pd.DataFrame({
                "Rank": ranks[:100],
                "Item": word_counts.index[:100],
                "Frequency": freqs[:100],
                "Expected": expected[:100].round(1),
            }),
            "interpretation": "Zipf's Law: Word frequency from titles is inversely proportional to rank."
        }
    
    def _analyze_price_law(self, entity):
        """Analyze Price's Law using biblium counts."""
        # Try to get counts from biblium
        if entity == "Authors":
            count_attr = 'authors_counts_df'
            if not hasattr(self.bib, count_attr) or self.bib.authors_counts_df is None:
                self.bib.count_authors()
            counts_df = getattr(self.bib, count_attr, None)
            name_col = 'Author'
        elif entity == "Sources":
            count_attr = 'sources_counts_df'
            if not hasattr(self.bib, count_attr) or self.bib.sources_counts_df is None:
                self.bib.count_sources()
            counts_df = getattr(self.bib, count_attr, None)
            name_col = 'Source'
        else:
            count_attr = 'author_keywords_counts_df'
            if not hasattr(self.bib, count_attr) or self.bib.author_keywords_counts_df is None:
                self.bib.count_author_keywords()
            counts_df = getattr(self.bib, count_attr, None)
            name_col = 'Keyword'
        
        if counts_df is not None and len(counts_df) > 0:
            # Find the document count column
            doc_col = None
            for col in counts_df.columns:
                if 'number' in col.lower() and 'document' in col.lower():
                    doc_col = col
                    break
            
            if doc_col:
                total_items = len(counts_df)
                total_pubs = counts_df[doc_col].sum()
                elite_n = int(np.sqrt(total_items))
                
                elite_pubs = counts_df.head(elite_n)[doc_col].sum()
                elite_pct = elite_pubs / total_pubs * 100
                
                return {
                    "law": "Price's Law (Elite Productivity)",
                    "entity": entity,
                    "data": pd.DataFrame({
                        "Metric": [f"Total {entity}", f"Elite (√n)", "Elite Publications", "Elite %", "Expected (50%)"],
                        "Value": [total_items, elite_n, int(elite_pubs), f"{elite_pct:.1f}%", "50%"],
                    }),
                    "interpretation": f"Price's Law: √n elite {entity.lower()} ({elite_n}) produce {elite_pct:.1f}% of output (expected ~50%).",
                    "from_biblium": True
                }
        
        # Fallback to manual calculation
        return self._fallback_price(entity)
    
    def _fallback_price(self, entity):
        """Fallback for Price's Law."""
        df = self.bib.df
        
        # Find appropriate column
        if entity == "Authors":
            col = None
            for c in ["Authors", "Authors or Inventors", "Author full names", "AU"]:
                if c in df.columns:
                    col = c
                    break
        elif entity == "Sources":
            col = None
            for c in ["Source title", "Source Title", "Journal", "Publication Name", "SO"]:
                if c in df.columns:
                    col = c
                    break
        else:
            col = None
            for c in ["Author Keywords", "Keywords", "DE", "ID"]:
                if c in df.columns:
                    col = c
                    break
        
        if col is None:
            raise ValueError(f"No {entity} column found")
        
        if entity in ["Authors", "Keywords"]:
            all_items = df[col].dropna().str.split(";").explode().str.strip()
        else:
            all_items = df[col].dropna()
        
        item_counts = all_items.value_counts()
        total_items = len(item_counts)
        total_pubs = item_counts.sum()
        elite_n = int(np.sqrt(total_items))
        
        sorted_counts = item_counts.sort_values(ascending=False)
        elite_pubs = sorted_counts.head(elite_n).sum()
        elite_pct = elite_pubs / total_pubs * 100
        
        return {
            "law": "Price's Law (Elite Productivity)",
            "entity": entity,
            "data": pd.DataFrame({
                "Metric": [f"Total {entity}", f"Elite (√n)", "Elite Publications", "Elite %", "Expected (50%)"],
                "Value": [total_items, elite_n, int(elite_pubs), f"{elite_pct:.1f}%", "50%"],
            }),
            "interpretation": f"Price's Law: √n elite {entity.lower()} ({elite_n}) produce {elite_pct:.1f}% of output (expected ~50%)."
        }
    
    def _analyze_pareto(self, entity):
        """Analyze Pareto Principle (80/20 rule) using biblium counts."""
        # Try to get counts from biblium
        if entity == "Authors":
            count_attr = 'authors_counts_df'
            if not hasattr(self.bib, count_attr) or self.bib.authors_counts_df is None:
                self.bib.count_authors()
            counts_df = getattr(self.bib, count_attr, None)
        elif entity == "Sources":
            count_attr = 'sources_counts_df'
            if not hasattr(self.bib, count_attr) or self.bib.sources_counts_df is None:
                self.bib.count_sources()
            counts_df = getattr(self.bib, count_attr, None)
        else:
            count_attr = 'author_keywords_counts_df'
            if not hasattr(self.bib, count_attr) or self.bib.author_keywords_counts_df is None:
                self.bib.count_author_keywords()
            counts_df = getattr(self.bib, count_attr, None)
        
        if counts_df is not None and len(counts_df) > 0:
            # Find the document count column
            doc_col = None
            for col in counts_df.columns:
                if 'number' in col.lower() and 'document' in col.lower():
                    doc_col = col
                    break
            
            if doc_col:
                total_items = len(counts_df)
                total_pubs = counts_df[doc_col].sum()
                
                top_20_n = max(1, int(total_items * 0.2))
                top_20_pubs = counts_df.head(top_20_n)[doc_col].sum()
                top_20_pct = top_20_pubs / total_pubs * 100
                
                return {
                    "law": "Pareto Principle (80/20 Rule)",
                    "entity": entity,
                    "data": pd.DataFrame({
                        "Metric": [f"Total {entity}", "Top 20%", "Their Publications", "Their %", "Expected (80%)"],
                        "Value": [total_items, top_20_n, int(top_20_pubs), f"{top_20_pct:.1f}%", "80%"],
                    }),
                    "interpretation": f"Pareto: Top 20% of {entity.lower()} ({top_20_n}) produce {top_20_pct:.1f}% of output (expected ~80%).",
                    "from_biblium": True
                }
        
        # Fallback
        return self._fallback_pareto(entity)
    
    def _fallback_pareto(self, entity):
        """Fallback for Pareto analysis."""
        df = self.bib.df
        
        # Find appropriate column
        if entity == "Authors":
            col = None
            for c in ["Authors", "Authors or Inventors", "Author full names", "AU"]:
                if c in df.columns:
                    col = c
                    break
        elif entity == "Sources":
            col = None
            for c in ["Source title", "Source Title", "Journal", "Publication Name", "SO"]:
                if c in df.columns:
                    col = c
                    break
        else:
            col = None
            for c in ["Author Keywords", "Keywords", "DE", "ID"]:
                if c in df.columns:
                    col = c
                    break
        
        if col is None:
            raise ValueError(f"No {entity} column found")
        
        if entity in ["Authors", "Keywords"]:
            all_items = df[col].dropna().str.split(";").explode().str.strip()
        else:
            all_items = df[col].dropna()
        
        item_counts = all_items.value_counts()
        total_items = len(item_counts)
        total_pubs = item_counts.sum()
        
        top_20_n = max(1, int(total_items * 0.2))
        sorted_counts = item_counts.sort_values(ascending=False)
        top_20_pubs = sorted_counts.head(top_20_n).sum()
        top_20_pct = top_20_pubs / total_pubs * 100
        
        return {
            "law": "Pareto Principle (80/20 Rule)",
            "entity": entity,
            "data": pd.DataFrame({
                "Metric": [f"Total {entity}", "Top 20%", "Their Publications", "Their %", "Expected (80%)"],
                "Value": [total_items, top_20_n, int(top_20_pubs), f"{top_20_pct:.1f}%", "80%"],
            }),
            "interpretation": f"Pareto: Top 20% of {entity.lower()} ({top_20_n}) produce {top_20_pct:.1f}% of output (expected ~80%)."
        }
    
    def _on_analysis_success(self, result: Dict):
        """Display analysis results."""
        self._show_results(result)
    
    def _on_analysis_error(self, error: str):
        """Handle analysis error."""
        self._show_error(f"Analysis error: {error}")
    
    def _widget_exists(self, widget) -> bool:
        """Check if a tkinter widget still exists."""
        if widget is None:
            return False
        try:
            return widget.winfo_exists()
        except (tk.TclError, AttributeError):
            return False
    
    def _show_results(self, result: Dict):
        """Display results."""
        # Check if results_tab still exists
        if not self._widget_exists(self.results_tab):
            # Recreate the results tab if it was destroyed
            try:
                if not self._widget_exists(self.results_notebook):
                    return
                self.results_tab = tk.Frame(self.results_notebook, bg=self.theme["bg_card"])
                self.results_notebook.insert(0, self.results_tab, text="📊 Results")
            except tk.TclError:
                # If even the notebook is gone, we can't show results
                return
        
        try:
            for widget in self.results_tab.winfo_children():
                widget.destroy()
        except tk.TclError:
            return
        
        # Interpretation
        if "interpretation" in result:
            tk.Label(
                self.results_tab, text=result["interpretation"],
                font=FONTS.get_font("body"),
                bg=self.theme["bg_card"], fg=self.theme["text_primary"],
                wraplength=500, justify=tk.LEFT,
            ).pack(fill=tk.X, pady=(0, 16))
        
        # Show stats if available (from biblium)
        if "stats_df" in result and result["stats_df"] is not None:
            stats_df = result["stats_df"]
            if len(stats_df) > 0:
                stats_frame = tk.Frame(self.results_tab, bg=self.theme["bg_card"])
                stats_frame.pack(fill=tk.X, pady=(0, 8))
                
                tk.Label(
                    stats_frame, text="Goodness of Fit:",
                    font=FONTS.get_font("body", bold=True),
                    bg=self.theme["bg_card"], fg=self.theme["text_primary"],
                ).pack(anchor=tk.W)
                
                for _, row in stats_df.iterrows():
                    measure = row.iloc[0] if len(row) > 0 else ""
                    value = row.iloc[1] if len(row) > 1 else ""
                    tk.Label(
                        stats_frame, text=f"  {measure}: {value:.4f}" if isinstance(value, (int, float)) else f"  {measure}: {value}",
                        font=FONTS.get_font("small"),
                        bg=self.theme["bg_card"], fg=self.theme["text_secondary"],
                    ).pack(anchor=tk.W)
        
        # Plot
        if HAS_MATPLOTLIB and "data" in result:
            plot = PlotFrame(self.results_tab, theme=self.theme_name, figsize=(8, 5), show_ai_button=True)
            plot.pack(fill=tk.X, pady=(0, 16))
            
            fig, ax = plot.get_figure()
            data = result["data"]
            law = result["law"]
            
            if "Lotka" in law:
                x_col = "Publications" if "Publications" in data.columns else data.columns[0]
                y_col = "Authors" if "Authors" in data.columns else data.columns[1]
                
                ax.scatter(data[x_col], data[y_col], color=self.theme["accent_primary"], label="Observed", s=50)
                
                if "Expected" in data.columns and self.show_fit_cb.get():
                    ax.plot(data[x_col], data["Expected"], 'r--', label="Theoretical", linewidth=2)
                
                if self.log_scale_cb.get():
                    ax.set_xscale('log')
                    ax.set_yscale('log')
                
                ax.set_xlabel("Publications")
                ax.set_ylabel("Number of Authors")
                
            elif "Zipf" in law:
                x_col = "Rank" if "Rank" in data.columns else data.columns[0]
                y_col = "Frequency" if "Frequency" in data.columns else data.columns[1]
                
                ax.scatter(data[x_col], data[y_col], color=self.theme["accent_primary"], label="Observed", s=20, alpha=0.7)
                
                if "Expected" in data.columns and self.show_fit_cb.get():
                    ax.plot(data[x_col], data["Expected"], 'r--', label="Theoretical", linewidth=2)
                
                if self.log_scale_cb.get():
                    ax.set_xscale('log')
                    ax.set_yscale('log')
                
                ax.set_xlabel("Rank")
                ax.set_ylabel("Frequency")
                
            elif "Bradford" in law:
                rank_col = "Rank" if "Rank" in data.columns else data.columns[0]
                cum_col = "Cumulative" if "Cumulative" in data.columns else "Cum %" if "Cum %" in data.columns else data.columns[3]
                
                ax.plot(data[rank_col], data[cum_col], color=self.theme["accent_primary"], linewidth=2)
                
                # Add zone markers if Zone column exists
                if "Zone" in data.columns:
                    for zone in data["Zone"].unique():
                        zone_data = data[data["Zone"] == zone]
                        if len(zone_data) > 0:
                            ax.axvline(x=zone_data[rank_col].iloc[-1], color='gray', linestyle='--', alpha=0.5)
                
                ax.set_xlabel("Source Rank")
                ax.set_ylabel("Cumulative Articles")
                
            elif "Price" in law or "Pareto" in law:
                # For summary tables, show a simple bar chart
                if "Metric" in data.columns:
                    metrics = data["Metric"].tolist()
                    values = []
                    for v in data["Value"]:
                        try:
                            if isinstance(v, str) and '%' in v:
                                values.append(float(v.replace('%', '')))
                            else:
                                values.append(float(v))
                        except:
                            values.append(0)
                    
                    # Only plot numeric values
                    plot_data = [(m, v) for m, v in zip(metrics, values) if v > 0 and '%' not in str(data[data["Metric"]==m]["Value"].values[0]) or 'Elite %' in m or 'Their %' in m]
                    if plot_data:
                        ax.barh([p[0] for p in plot_data], [p[1] for p in plot_data], color=self.theme["accent_primary"])
                else:
                    ax.bar(range(len(data)), data.iloc[:, 1], color=self.theme["accent_primary"])
            
            ax.legend()
            ax.set_title(law.split("(")[0].strip())
            fig.tight_layout()
            plot.refresh()
        
        # Table
        if "data" in result:
            table = DataTable(self.results_tab, theme=self.theme_name, max_rows=50)
            table.pack(fill=tk.BOTH, expand=True)
            table.set_data(result["data"])



# -*- coding: utf-8 -*-
"""
Disruption Index Module for Biblium
===================================

This module implements the Disruption Index (CD Index) and related metrics
for measuring whether research consolidates or disrupts existing knowledge.

The Disruption Index was introduced by Funk & Owen-Smith (2017) and refined
by Wu, Wang & Evans (2019) in "Large teams develop and small teams disrupt 
science and technology" (Nature).

Key metrics:
- CD Index: Core disruption index ranging from -1 (consolidating) to +1 (disruptive)
- DI Index: Simplified disruption index
- CD5 Index: 5-year window disruption index

Aggregation levels:
- Document-level: Individual paper disruption
- Source-level: Average disruption by journal/source
- Author-level: Average disruption by author
- Country-level: Average disruption by country
- Institution-level: Average disruption by affiliation

@author: Claude (Anthropic) for Lan.Umek
@version: 2.7.0
"""

from __future__ import annotations

import os
import warnings
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd

try:
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches
    from matplotlib.colors import TwoSlopeNorm
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False

try:
    import seaborn as sns
    HAS_SEABORN = True
except ImportError:
    HAS_SEABORN = False


# =============================================================================
# CITATION NETWORK BUILDING
# =============================================================================

def build_citation_network_from_refs(
    df: pd.DataFrame,
    id_col: str = "unique-id",
    refs_col: str = "References",
    sep: str = "; ",
    normalize_ids: bool = True,
) -> Dict[str, Set[str]]:
    """
    Build citation network from references column.
    
    Parameters
    ----------
    df : pd.DataFrame
        Bibliographic dataframe.
    id_col : str
        Column containing document IDs.
    refs_col : str
        Column containing references (semicolon-separated).
    sep : str
        Separator for parsing references.
    normalize_ids : bool
        Whether to normalize IDs (lowercase, strip).
    
    Returns
    -------
    Dict[str, Set[str]]
        Dictionary mapping paper_id -> set of paper_ids it cites.
    """
    network = {}
    
    for _, row in df.iterrows():
        doc_id = str(row.get(id_col, ""))
        if not doc_id or doc_id == "nan":
            continue
            
        if normalize_ids:
            doc_id = doc_id.strip().lower()
        
        refs_str = row.get(refs_col, "")
        if pd.isna(refs_str) or not refs_str:
            network[doc_id] = set()
            continue
        
        refs = set()
        for ref in str(refs_str).split(sep):
            ref = ref.strip()
            if ref and ref != "nan":
                if normalize_ids:
                    ref = ref.lower()
                refs.add(ref)
        
        network[doc_id] = refs
    
    return network


def build_reverse_citation_index(
    citation_network: Dict[str, Set[str]]
) -> Dict[str, Set[str]]:
    """
    Build reverse citation index (who cites whom).
    
    Parameters
    ----------
    citation_network : Dict[str, Set[str]]
        Forward citation network {paper_id: set of refs}.
    
    Returns
    -------
    Dict[str, Set[str]]
        Reverse index {paper_id: set of papers that cite it}.
    """
    cited_by = defaultdict(set)
    
    for paper, refs in citation_network.items():
        for ref in refs:
            cited_by[ref].add(paper)
    
    return dict(cited_by)


# =============================================================================
# CORE DISRUPTION INDEX COMPUTATION
# =============================================================================

def compute_disruption_for_paper(
    doc_id: str,
    focal_refs: Set[str],
    citing_papers: Set[str],
    citation_network: Dict[str, Set[str]],
    cited_by: Dict[str, Set[str]],
) -> Dict[str, Any]:
    """
    Compute disruption metrics for a single paper.
    
    The CD Index formula: CD = (n_i - n_j) / (n_i + n_j + n_k)
    
    Where:
    - n_i = citing papers that cite focal but NOT its references
    - n_j = citing papers that cite BOTH focal AND its references  
    - n_k = papers that cite focal's references but NOT focal
    
    Parameters
    ----------
    doc_id : str
        Document identifier.
    focal_refs : Set[str]
        References of the focal paper.
    citing_papers : Set[str]
        Papers that cite the focal paper.
    citation_network : Dict[str, Set[str]]
        Forward citation network.
    cited_by : Dict[str, Set[str]]
        Reverse citation index.
    
    Returns
    -------
    Dict with disruption metrics.
    """
    if not citing_papers:
        return {
            'doc_id': doc_id,
            'cd_index': np.nan,
            'di_index': np.nan,
            'n_citing': 0,
            'n_i': 0,
            'n_j': 0,
            'n_k': 0,
            'interpretation': 'uncited',
        }
    
    # Count n_i, n_j for citing papers
    n_i = 0  # Cite focal, don't cite focal's refs
    n_j = 0  # Cite focal AND focal's refs
    
    for citing_paper in citing_papers:
        citing_refs = citation_network.get(citing_paper, set())
        
        # Does this citing paper also cite any of focal's references?
        cites_focal_refs = bool(citing_refs & focal_refs)
        
        if cites_focal_refs:
            n_j += 1
        else:
            n_i += 1
    
    # Count n_k: papers that cite focal's references but not focal
    papers_citing_refs = set()
    for ref in focal_refs:
        ref_citers = cited_by.get(ref, set())
        papers_citing_refs.update(ref_citers)
    
    # Exclude papers that cite focal and focal itself
    n_k = len(papers_citing_refs - citing_papers - {doc_id})
    
    # Compute CD index
    denominator = n_i + n_j + n_k
    if denominator > 0:
        cd_index = (n_i - n_j) / denominator
    else:
        cd_index = 0.0
    
    # Simplified DI index (only considers citing papers)
    if n_i + n_j > 0:
        di_index = (n_i - n_j) / (n_i + n_j)
    else:
        di_index = 0.0
    
    # Interpret
    if cd_index >= 0.25:
        interpretation = "disruptive"
    elif cd_index <= -0.25:
        interpretation = "consolidating"
    else:
        interpretation = "neutral"
    
    return {
        'doc_id': doc_id,
        'cd_index': round(cd_index, 4),
        'di_index': round(di_index, 4),
        'n_citing': len(citing_papers),
        'n_i': n_i,
        'n_j': n_j,
        'n_k': n_k,
        'interpretation': interpretation,
    }


def compute_document_disruption(
    df: pd.DataFrame,
    id_col: str = "unique-id",
    refs_col: str = "References",
    sep: str = "; ",
    citation_network: Dict[str, Set[str]] = None,
    verbose: bool = True,
) -> pd.DataFrame:
    """
    Compute Disruption Index for all documents.
    
    Parameters
    ----------
    df : pd.DataFrame
        Bibliographic dataframe.
    id_col : str
        Column with document IDs.
    refs_col : str
        Column with references.
    sep : str
        Separator for parsing references.
    citation_network : Dict[str, Set[str]], optional
        Pre-computed citation network.
    verbose : bool
        Print progress.
    
    Returns
    -------
    pd.DataFrame
        DataFrame with disruption metrics for each document.
    """
    if verbose:
        print("Computing Document Disruption Index...")
    
    # Build citation network if not provided
    if citation_network is None:
        if verbose:
            print("  Building citation network from references...")
        citation_network = build_citation_network_from_refs(
            df, id_col, refs_col, sep
        )
    
    # Build reverse citation index
    if verbose:
        print("  Building reverse citation index...")
    cited_by = build_reverse_citation_index(citation_network)
    
    # Get all paper IDs in dataset
    all_papers = set()
    for doc_id in df[id_col].dropna():
        all_papers.add(str(doc_id).strip().lower())
    
    # Diagnostic info
    internal_citations = sum(
        1 for paper in all_papers 
        if paper in cited_by and cited_by[paper] & all_papers
    )
    
    if verbose:
        print(f"  Papers in dataset: {len(all_papers)}")
        print(f"  Papers with internal citations: {internal_citations}")
        
        if internal_citations < len(all_papers) * 0.1:
            print("  NOTE: Few internal citations detected.")
            print("    CD Index requires papers to cite each other within the dataset.")
            print("    Papers citing only external sources will have CD = NaN.")
    
    # Compute metrics for each paper
    results = []
    
    for _, row in df.iterrows():
        doc_id = str(row.get(id_col, ""))
        if not doc_id or doc_id == "nan":
            continue
        
        doc_id_norm = doc_id.strip().lower()
        
        # Get focal paper's references
        focal_refs = citation_network.get(doc_id_norm, set())
        
        # Get papers that cite the focal paper (internal only)
        citing_papers = cited_by.get(doc_id_norm, set()) & all_papers
        
        # Compute disruption
        metrics = compute_disruption_for_paper(
            doc_id_norm, focal_refs, citing_papers,
            citation_network, cited_by
        )
        
        # Use original doc_id for output
        metrics['doc_id'] = doc_id
        results.append(metrics)
    
    if verbose:
        print(f"  Computed disruption for {len(results)} documents")
    
    return pd.DataFrame(results)


# =============================================================================
# AGGREGATED DISRUPTION (SOURCES, AUTHORS, COUNTRIES, ETC.)
# =============================================================================

def aggregate_disruption_by_entity(
    disruption_df: pd.DataFrame,
    main_df: pd.DataFrame,
    entity_col: str,
    id_col: str = "unique-id",
    sep: str = "; ",
    min_docs: int = 1,
    metrics: List[str] = None,
) -> pd.DataFrame:
    """
    Aggregate disruption metrics by entity (source, author, country, etc.).
    
    Parameters
    ----------
    disruption_df : pd.DataFrame
        Document-level disruption metrics.
    main_df : pd.DataFrame
        Main bibliographic dataframe with entity column.
    entity_col : str
        Column to aggregate by.
    id_col : str
        Column with document IDs.
    sep : str
        Separator for multi-value columns.
    min_docs : int
        Minimum documents for entity to be included.
    metrics : List[str], optional
        Which metrics to aggregate. Default: cd_index, di_index.
    
    Returns
    -------
    pd.DataFrame
        Entity-level disruption statistics.
    """
    if metrics is None:
        metrics = ['cd_index', 'di_index']
    
    # Start with only the columns we need from main_df to avoid conflicts
    needed_cols = [id_col, entity_col]
    needed_cols = [c for c in needed_cols if c in main_df.columns]
    merged = main_df[needed_cols].copy()
    
    # Normalize doc_id for joining
    merged['_doc_id_norm'] = merged[id_col].astype(str).str.strip().str.lower()
    disruption_copy = disruption_df.copy()
    disruption_copy['_doc_id_norm'] = disruption_copy['doc_id'].astype(str).str.strip().str.lower()
    
    # Get columns to merge from disruption_df
    cols_to_merge = ['_doc_id_norm'] + [c for c in metrics + ['interpretation', 'n_citing'] 
                                         if c in disruption_copy.columns]
    
    merged = merged.merge(
        disruption_copy[cols_to_merge],
        on='_doc_id_norm',
        how='left'
    )
    
    # Handle multi-value columns (e.g., Authors)
    if entity_col not in merged.columns:
        raise ValueError(f"Column '{entity_col}' not found in dataframe")
    
    # Check if column contains separators (only for string columns)
    sample = merged[entity_col].dropna().head(100)
    is_multi = False
    if sample.dtype == 'object' or str(sample.dtype) == 'string':
        try:
            is_multi = sample.str.contains(sep, regex=False).any() if len(sample) > 0 else False
        except (AttributeError, TypeError):
            is_multi = False
    
    if is_multi:
        # Explode multi-value column
        merged['_entity_list'] = merged[entity_col].fillna('').str.split(sep)
        merged = merged.explode('_entity_list')
        merged['_entity_list'] = merged['_entity_list'].str.strip()
        merged = merged[merged['_entity_list'] != '']
        entity_col_use = '_entity_list'
    else:
        entity_col_use = entity_col
    
    # Build aggregation dictionary
    agg_dict = {}
    
    # Count documents
    agg_dict['_doc_id_norm'] = 'count'
    
    # Add metrics that exist in the merged dataframe
    metrics_present = [m for m in metrics if m in merged.columns]
    for metric in metrics_present:
        agg_dict[metric] = ['mean', 'median', 'std', 'min', 'max']
    
    if 'n_citing' in merged.columns:
        agg_dict['n_citing'] = 'sum'
    
    # Group and aggregate
    grouped = merged.groupby(entity_col_use).agg(agg_dict)
    
    # Flatten column names - handle both MultiIndex and regular Index
    if isinstance(grouped.columns, pd.MultiIndex):
        new_columns = []
        for col in grouped.columns:
            if col[1]:
                new_columns.append(f'{col[0]}_{col[1]}')
            else:
                new_columns.append(col[0])
        grouped.columns = new_columns
    else:
        # Single aggregation per column - rename directly
        rename_map = {}
        for col in grouped.columns:
            if col == '_doc_id_norm':
                rename_map[col] = 'n_documents'
            elif col == 'n_citing':
                rename_map[col] = 'total_citations'
        grouped = grouped.rename(columns=rename_map)
    
    # Rename count column (for MultiIndex case)
    if '_doc_id_norm_count' in grouped.columns:
        grouped = grouped.rename(columns={'_doc_id_norm_count': 'n_documents'})
    if 'n_citing_sum' in grouped.columns:
        grouped = grouped.rename(columns={'n_citing_sum': 'total_citations'})
    
    # Filter by min_docs
    grouped = grouped[grouped['n_documents'] >= min_docs]
    
    # Sort by mean cd_index
    if 'cd_index_mean' in grouped.columns:
        grouped = grouped.sort_values('cd_index_mean', ascending=False)
    
    # Reset index
    grouped = grouped.reset_index()
    grouped = grouped.rename(columns={entity_col_use: entity_col})
    
    # Add interpretation counts
    if 'interpretation' in merged.columns:
        interp_counts = merged.groupby(entity_col_use)['interpretation'].value_counts().unstack(fill_value=0)
        interp_counts = interp_counts.reset_index()
        interp_counts = interp_counts.rename(columns={entity_col_use: entity_col})
        
        # Merge interpretation counts
        for col in ['disruptive', 'neutral', 'consolidating', 'uncited']:
            if col in interp_counts.columns:
                grouped = grouped.merge(
                    interp_counts[[entity_col, col]],
                    on=entity_col,
                    how='left'
                )
                grouped[f'n_{col}'] = grouped[col].fillna(0).astype(int)
                grouped = grouped.drop(columns=[col])
    
    return grouped


def compute_source_disruption(
    disruption_df: pd.DataFrame,
    main_df: pd.DataFrame,
    source_col: str = "Source title",
    id_col: str = "unique-id",
    min_docs: int = 3,
) -> pd.DataFrame:
    """
    Compute average Disruption Index by source/journal.
    
    Parameters
    ----------
    disruption_df : pd.DataFrame
        Document-level disruption metrics.
    main_df : pd.DataFrame
        Main bibliographic dataframe.
    source_col : str
        Column containing source/journal names.
    id_col : str
        Column with document IDs.
    min_docs : int
        Minimum documents for source to be included.
    
    Returns
    -------
    pd.DataFrame
        Source-level disruption statistics.
    """
    return aggregate_disruption_by_entity(
        disruption_df, main_df, source_col, id_col,
        sep="; ", min_docs=min_docs
    )


def compute_author_disruption(
    disruption_df: pd.DataFrame,
    main_df: pd.DataFrame,
    author_col: str = "Authors",
    id_col: str = "unique-id",
    sep: str = "; ",
    min_docs: int = 2,
) -> pd.DataFrame:
    """
    Compute average Disruption Index by author.
    
    Parameters
    ----------
    disruption_df : pd.DataFrame
        Document-level disruption metrics.
    main_df : pd.DataFrame
        Main bibliographic dataframe.
    author_col : str
        Column containing author names.
    id_col : str
        Column with document IDs.
    sep : str
        Separator for author names.
    min_docs : int
        Minimum documents for author to be included.
    
    Returns
    -------
    pd.DataFrame
        Author-level disruption statistics.
    """
    return aggregate_disruption_by_entity(
        disruption_df, main_df, author_col, id_col,
        sep=sep, min_docs=min_docs
    )


def compute_country_disruption(
    disruption_df: pd.DataFrame,
    main_df: pd.DataFrame,
    country_col: str = "Countries",
    id_col: str = "unique-id",
    sep: str = "; ",
    min_docs: int = 5,
) -> pd.DataFrame:
    """
    Compute average Disruption Index by country.
    
    Parameters
    ----------
    disruption_df : pd.DataFrame
        Document-level disruption metrics.
    main_df : pd.DataFrame
        Main bibliographic dataframe.
    country_col : str
        Column containing country names.
    id_col : str
        Column with document IDs.
    sep : str
        Separator for country names.
    min_docs : int
        Minimum documents for country to be included.
    
    Returns
    -------
    pd.DataFrame
        Country-level disruption statistics.
    """
    return aggregate_disruption_by_entity(
        disruption_df, main_df, country_col, id_col,
        sep=sep, min_docs=min_docs
    )


def compute_affiliation_disruption(
    disruption_df: pd.DataFrame,
    main_df: pd.DataFrame,
    affiliation_col: str = "Affiliations",
    id_col: str = "unique-id",
    sep: str = "; ",
    min_docs: int = 3,
) -> pd.DataFrame:
    """
    Compute average Disruption Index by affiliation/institution.
    
    Parameters
    ----------
    disruption_df : pd.DataFrame
        Document-level disruption metrics.
    main_df : pd.DataFrame
        Main bibliographic dataframe.
    affiliation_col : str
        Column containing affiliation names.
    id_col : str
        Column with document IDs.
    sep : str
        Separator for affiliation names.
    min_docs : int
        Minimum documents for affiliation to be included.
    
    Returns
    -------
    pd.DataFrame
        Affiliation-level disruption statistics.
    """
    return aggregate_disruption_by_entity(
        disruption_df, main_df, affiliation_col, id_col,
        sep=sep, min_docs=min_docs
    )


def compute_year_disruption(
    disruption_df: pd.DataFrame,
    main_df: pd.DataFrame,
    year_col: str = "Year",
    id_col: str = "unique-id",
) -> pd.DataFrame:
    """
    Compute average Disruption Index by year.
    
    Parameters
    ----------
    disruption_df : pd.DataFrame
        Document-level disruption metrics.
    main_df : pd.DataFrame
        Main bibliographic dataframe.
    year_col : str
        Column containing publication year.
    id_col : str
        Column with document IDs.
    
    Returns
    -------
    pd.DataFrame
        Year-level disruption statistics.
    """
    return aggregate_disruption_by_entity(
        disruption_df, main_df, year_col, id_col,
        sep="|||", min_docs=1  # No separator for year
    )


# =============================================================================
# ADD DISRUPTION TO EXISTING DATAFRAMES
# =============================================================================

def add_disruption_to_df(
    df: pd.DataFrame,
    disruption_df: pd.DataFrame,
    id_col: str = "unique-id",
    columns: List[str] = None,
) -> pd.DataFrame:
    """
    Add disruption columns to an existing dataframe.
    
    Parameters
    ----------
    df : pd.DataFrame
        Target dataframe.
    disruption_df : pd.DataFrame
        Document-level disruption metrics.
    id_col : str
        Column with document IDs.
    columns : List[str], optional
        Which disruption columns to add. Default: all.
    
    Returns
    -------
    pd.DataFrame
        Dataframe with disruption columns added.
    """
    if columns is None:
        columns = ['cd_index', 'di_index', 'interpretation', 'n_citing', 'n_i', 'n_j', 'n_k']
    
    # Keep only requested columns
    cols_to_add = ['doc_id'] + [c for c in columns if c in disruption_df.columns]
    disruption_subset = disruption_df[cols_to_add].copy()
    
    # Normalize for joining
    df = df.copy()
    df['_doc_id_join'] = df[id_col].astype(str).str.strip().str.lower()
    disruption_subset['_doc_id_join'] = disruption_subset['doc_id'].astype(str).str.strip().str.lower()
    
    # Merge
    result = df.merge(
        disruption_subset.drop(columns=['doc_id']),
        on='_doc_id_join',
        how='left'
    )
    
    # Clean up
    result = result.drop(columns=['_doc_id_join'])
    
    return result


# =============================================================================
# VISUALIZATION FUNCTIONS
# =============================================================================

def plot_disruption_distribution(
    disruption_df: pd.DataFrame,
    metric: str = 'cd_index',
    title: str = None,
    figsize: Tuple[int, int] = (10, 6),
    color: str = None,
    ax: plt.Axes = None,
    show_stats: bool = True,
) -> plt.Figure:
    """
    Plot distribution of disruption index.
    
    Parameters
    ----------
    disruption_df : pd.DataFrame
        Disruption metrics dataframe.
    metric : str
        Which metric to plot.
    title : str, optional
        Plot title.
    figsize : tuple
        Figure size.
    color : str, optional
        Bar color.
    ax : plt.Axes, optional
        Existing axes to plot on.
    show_stats : bool
        Show statistics box.
    
    Returns
    -------
    plt.Figure
    """
    if not HAS_MATPLOTLIB:
        raise ImportError("matplotlib is required for plotting")
    
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.figure
    
    # Get valid data
    data = disruption_df[metric].dropna()
    
    if len(data) == 0:
        ax.text(0.5, 0.5, "No valid data", ha='center', va='center', transform=ax.transAxes)
        return fig
    
    # Histogram
    bins = np.linspace(-1, 1, 41)
    n, bins_edges, patches = ax.hist(data, bins=bins, color=color or '#3498db', 
                                      alpha=0.7, edgecolor='white', linewidth=0.5)
    
    # Color bars based on value
    for i, patch in enumerate(patches):
        bin_center = (bins_edges[i] + bins_edges[i+1]) / 2
        if bin_center > 0.25:
            patch.set_facecolor('#27ae60')  # Green for disruptive
        elif bin_center < -0.25:
            patch.set_facecolor('#e74c3c')  # Red for consolidating
        else:
            patch.set_facecolor('#95a5a6')  # Gray for neutral
    
    # Reference lines
    ax.axvline(x=0, color='black', linestyle='-', linewidth=1, alpha=0.5)
    ax.axvline(x=0.25, color='green', linestyle='--', linewidth=1, alpha=0.5)
    ax.axvline(x=-0.25, color='red', linestyle='--', linewidth=1, alpha=0.5)
    
    # Labels
    ax.set_xlabel(f'{metric.replace("_", " ").title()}', fontsize=11)
    ax.set_ylabel('Number of Documents', fontsize=11)
    ax.set_title(title or f'Distribution of {metric.replace("_", " ").title()}', fontsize=13)
    ax.set_xlim(-1.05, 1.05)
    
    # Statistics box
    if show_stats:
        stats_text = (
            f'n = {len(data):,}\n'
            f'Mean = {data.mean():.3f}\n'
            f'Median = {data.median():.3f}\n'
            f'Std = {data.std():.3f}'
        )
        ax.text(0.02, 0.98, stats_text, transform=ax.transAxes, fontsize=9,
                verticalalignment='top', bbox=dict(boxstyle='round', facecolor='white', alpha=0.8))
    
    # Legend
    legend_elements = [
        mpatches.Patch(facecolor='#27ae60', label='Disruptive (> 0.25)'),
        mpatches.Patch(facecolor='#95a5a6', label='Neutral'),
        mpatches.Patch(facecolor='#e74c3c', label='Consolidating (< -0.25)'),
    ]
    ax.legend(handles=legend_elements, loc='upper right', fontsize=9)
    
    plt.tight_layout()
    return fig


def plot_disruption_by_entity(
    entity_df: pd.DataFrame,
    entity_col: str,
    metric: str = 'cd_index_mean',
    top_n: int = 20,
    title: str = None,
    figsize: Tuple[int, int] = (12, 8),
    ax: plt.Axes = None,
    show_error: bool = True,
) -> plt.Figure:
    """
    Plot disruption index by entity (bar chart).
    
    Parameters
    ----------
    entity_df : pd.DataFrame
        Entity-level disruption statistics.
    entity_col : str
        Column with entity names.
    metric : str
        Which metric to plot.
    top_n : int
        Number of entities to show.
    title : str, optional
        Plot title.
    figsize : tuple
        Figure size.
    ax : plt.Axes, optional
        Existing axes.
    show_error : bool
        Show error bars (std).
    
    Returns
    -------
    plt.Figure
    """
    if not HAS_MATPLOTLIB:
        raise ImportError("matplotlib is required for plotting")
    
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.figure
    
    # Sort and get top N
    if metric in entity_df.columns:
        df_sorted = entity_df.nlargest(top_n, metric).copy()
    else:
        df_sorted = entity_df.head(top_n).copy()
    
    # Truncate long names
    df_sorted[entity_col] = df_sorted[entity_col].astype(str).str[:50]
    
    # Colors based on value
    colors = []
    for val in df_sorted[metric]:
        if val > 0.25:
            colors.append('#27ae60')
        elif val < -0.25:
            colors.append('#e74c3c')
        else:
            colors.append('#3498db')
    
    # Bar plot
    y_pos = range(len(df_sorted))
    bars = ax.barh(y_pos, df_sorted[metric], color=colors, alpha=0.8, edgecolor='white')
    
    # Error bars
    std_col = metric.replace('mean', 'std')
    if show_error and std_col in df_sorted.columns:
        ax.errorbar(df_sorted[metric], y_pos, xerr=df_sorted[std_col],
                   fmt='none', ecolor='gray', alpha=0.5, capsize=3)
    
    # Labels
    ax.set_yticks(y_pos)
    ax.set_yticklabels(df_sorted[entity_col], fontsize=9)
    ax.set_xlabel(metric.replace('_', ' ').title(), fontsize=11)
    ax.set_title(title or f'Disruption Index by {entity_col}', fontsize=13)
    
    # Reference line at 0
    ax.axvline(x=0, color='black', linestyle='-', linewidth=1, alpha=0.5)
    
    # Add document count as text
    if 'n_documents' in df_sorted.columns:
        for i, (idx, row) in enumerate(df_sorted.iterrows()):
            ax.text(ax.get_xlim()[1], i, f'  n={int(row["n_documents"])}',
                   va='center', fontsize=8, color='gray')
    
    ax.invert_yaxis()
    plt.tight_layout()
    return fig


def plot_disruption_over_time(
    year_df: pd.DataFrame,
    year_col: str = "Year",
    metric: str = 'cd_index_mean',
    title: str = None,
    figsize: Tuple[int, int] = (12, 6),
    ax: plt.Axes = None,
    show_ci: bool = True,
) -> plt.Figure:
    """
    Plot disruption index trend over time.
    
    Parameters
    ----------
    year_df : pd.DataFrame
        Year-level disruption statistics.
    year_col : str
        Column with years.
    metric : str
        Which metric to plot.
    title : str, optional
        Plot title.
    figsize : tuple
        Figure size.
    ax : plt.Axes, optional
        Existing axes.
    show_ci : bool
        Show confidence interval (±1 std).
    
    Returns
    -------
    plt.Figure
    """
    if not HAS_MATPLOTLIB:
        raise ImportError("matplotlib is required for plotting")
    
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.figure
    
    # Sort by year
    df_sorted = year_df.sort_values(year_col).copy()
    
    years = df_sorted[year_col].astype(int)
    values = df_sorted[metric]
    
    # Line plot
    ax.plot(years, values, 'o-', color='#3498db', linewidth=2, markersize=6, label='Mean CD Index')
    
    # Confidence interval
    std_col = metric.replace('mean', 'std')
    if show_ci and std_col in df_sorted.columns:
        ax.fill_between(years, 
                       values - df_sorted[std_col],
                       values + df_sorted[std_col],
                       alpha=0.2, color='#3498db', label='±1 Std')
    
    # Reference lines
    ax.axhline(y=0, color='black', linestyle='-', linewidth=1, alpha=0.5)
    ax.axhline(y=0.25, color='green', linestyle='--', linewidth=1, alpha=0.3, label='Disruptive threshold')
    ax.axhline(y=-0.25, color='red', linestyle='--', linewidth=1, alpha=0.3, label='Consolidating threshold')
    
    # Labels
    ax.set_xlabel('Year', fontsize=11)
    ax.set_ylabel(metric.replace('_', ' ').title(), fontsize=11)
    ax.set_title(title or 'Disruption Index Over Time', fontsize=13)
    ax.legend(loc='best', fontsize=9)
    
    # Set y limits
    ax.set_ylim(-1, 1)
    
    plt.tight_layout()
    return fig


def plot_disruption_scatter(
    disruption_df: pd.DataFrame,
    main_df: pd.DataFrame,
    id_col: str = "unique-id",
    x_col: str = "Year",
    y_col: str = "cd_index",
    size_col: str = "Cited by",
    color_by: str = "interpretation",
    title: str = None,
    figsize: Tuple[int, int] = (12, 8),
    ax: plt.Axes = None,
) -> plt.Figure:
    """
    Scatter plot of disruption vs. another variable.
    
    Parameters
    ----------
    disruption_df : pd.DataFrame
        Disruption metrics.
    main_df : pd.DataFrame
        Main bibliographic dataframe.
    id_col : str
        Document ID column.
    x_col : str
        Column for x-axis.
    y_col : str
        Column for y-axis (from disruption_df).
    size_col : str
        Column for point size.
    color_by : str
        Column to color by.
    title : str, optional
        Plot title.
    figsize : tuple
        Figure size.
    ax : plt.Axes, optional
        Existing axes.
    
    Returns
    -------
    plt.Figure
    """
    if not HAS_MATPLOTLIB:
        raise ImportError("matplotlib is required for plotting")
    
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.figure
    
    # Merge dataframes
    merged = add_disruption_to_df(main_df, disruption_df, id_col)
    
    # Get valid data
    if x_col not in merged.columns or y_col not in merged.columns:
        ax.text(0.5, 0.5, f"Column not found: {x_col} or {y_col}", 
               ha='center', va='center', transform=ax.transAxes)
        return fig
    
    data = merged[[x_col, y_col, color_by]].dropna()
    
    # Size
    if size_col in merged.columns:
        sizes = merged.loc[data.index, size_col].fillna(10)
        sizes = np.clip(sizes, 10, 1000)
        sizes = (sizes - sizes.min()) / (sizes.max() - sizes.min() + 1) * 200 + 20
    else:
        sizes = 50
    
    # Colors
    color_map = {
        'disruptive': '#27ae60',
        'neutral': '#95a5a6',
        'consolidating': '#e74c3c',
        'uncited': '#bdc3c7',
    }
    colors = data[color_by].map(color_map).fillna('#95a5a6')
    
    # Scatter
    ax.scatter(data[x_col], data[y_col], c=colors, s=sizes, alpha=0.6, edgecolor='white')
    
    # Reference lines
    ax.axhline(y=0, color='black', linestyle='-', linewidth=1, alpha=0.5)
    ax.axhline(y=0.25, color='green', linestyle='--', linewidth=1, alpha=0.3)
    ax.axhline(y=-0.25, color='red', linestyle='--', linewidth=1, alpha=0.3)
    
    # Labels
    ax.set_xlabel(x_col, fontsize=11)
    ax.set_ylabel(y_col.replace('_', ' ').title(), fontsize=11)
    ax.set_title(title or f'{y_col.replace("_", " ").title()} vs {x_col}', fontsize=13)
    
    # Legend
    legend_elements = [mpatches.Patch(facecolor=c, label=l) 
                       for l, c in color_map.items() if l in data[color_by].values]
    ax.legend(handles=legend_elements, loc='best', fontsize=9)
    
    plt.tight_layout()
    return fig


def plot_disruption_comparison(
    entity_dfs: Dict[str, pd.DataFrame],
    metric: str = 'cd_index_mean',
    title: str = None,
    figsize: Tuple[int, int] = (10, 6),
) -> plt.Figure:
    """
    Compare disruption distributions across entity types.
    
    Parameters
    ----------
    entity_dfs : Dict[str, pd.DataFrame]
        Dictionary mapping entity type to disruption dataframe.
    metric : str
        Which metric to compare.
    title : str, optional
        Plot title.
    figsize : tuple
        Figure size.
    
    Returns
    -------
    plt.Figure
    """
    if not HAS_MATPLOTLIB:
        raise ImportError("matplotlib is required for plotting")
    
    fig, ax = plt.subplots(figsize=figsize)
    
    data_to_plot = []
    labels = []
    
    for entity_type, df in entity_dfs.items():
        if metric in df.columns:
            data_to_plot.append(df[metric].dropna().values)
            labels.append(entity_type)
    
    if not data_to_plot:
        ax.text(0.5, 0.5, "No data to plot", ha='center', va='center', transform=ax.transAxes)
        return fig
    
    # Box plot
    bp = ax.boxplot(data_to_plot, labels=labels, patch_artist=True)
    
    colors = ['#3498db', '#e74c3c', '#27ae60', '#9b59b6', '#f39c12']
    for patch, color in zip(bp['boxes'], colors[:len(bp['boxes'])]):
        patch.set_facecolor(color)
        patch.set_alpha(0.7)
    
    # Reference lines
    ax.axhline(y=0, color='black', linestyle='-', linewidth=1, alpha=0.5)
    ax.axhline(y=0.25, color='green', linestyle='--', linewidth=1, alpha=0.3)
    ax.axhline(y=-0.25, color='red', linestyle='--', linewidth=1, alpha=0.3)
    
    ax.set_ylabel(metric.replace('_', ' ').title(), fontsize=11)
    ax.set_title(title or 'Disruption Index Comparison', fontsize=13)
    
    plt.tight_layout()
    return fig


# =============================================================================
# MIXIN CLASS FOR INTEGRATION WITH BIBLIOSTATS
# =============================================================================

class DisruptionMixin:
    """
    Mixin class providing disruption index methods for BiblioStats.
    
    This mixin adds methods to compute disruption metrics and adds them
    to the main df and entity count dataframes.
    """
    
    # Storage attributes
    disruption_df: pd.DataFrame = None
    
    def compute_disruption_index(
        self,
        id_col: str = None,
        refs_col: str = None,
        sep: str = None,
        verbose: bool = True,
        add_to_df: bool = True,
        add_to_counts: bool = True,
    ) -> pd.DataFrame:
        """
        Compute Disruption Index (CD Index) for all documents.
        
        The CD Index measures whether a paper consolidates or disrupts existing
        knowledge:
        - CD = +1: Fully disruptive (citers ignore the paper's references)
        - CD = 0: Neutral  
        - CD = -1: Fully consolidating (citers always cite the paper's references)
        
        Parameters
        ----------
        id_col : str, optional
            Column with document IDs. Auto-detected if not provided.
        refs_col : str, optional
            Column with references. Auto-detected if not provided.
        sep : str, optional
            Separator for references. Default: "; "
        verbose : bool
            Print progress messages.
        add_to_df : bool
            Add disruption columns to self.df.
        add_to_counts : bool
            Add disruption columns to entity count dataframes.
        
        Returns
        -------
        pd.DataFrame
            Document-level disruption metrics.
        
        Notes
        -----
        The CD Index is computed based on citations within the dataset. This works
        best when:
        
        - **OpenAlex**: References contain OpenAlex IDs that can be directly matched.
          This provides the most accurate results.
        
        - **Scopus/Web of Science**: References are formatted citation strings.
          The algorithm attempts to match references to documents in the dataset
          by comparing IDs (DOI, EID, UT). Results depend on how many documents
          in your dataset cite each other.
        
        For best results with Scopus/WoS data, use a focused dataset where papers
        are likely to cite each other (e.g., papers from a specific research area
        or journal over time).
        
        Examples
        --------
        >>> bib = BiblioAnalysis("data.csv", db="scopus")
        >>> di = bib.compute_disruption_index()
        >>> print(di[['doc_id', 'cd_index', 'interpretation']].head())
        """
        # Auto-detect columns
        if id_col is None:
            id_col = self.mapping.get("unique-id", "unique-id")
            if id_col not in self.df.columns:
                for alt in ["DOI", "EID", "unique-id", "UT"]:
                    if alt in self.df.columns:
                        id_col = alt
                        break
        
        if refs_col is None:
            refs_col = self.mapping.get("References", "References")
            if refs_col not in self.df.columns:
                for alt in ["References", "Cited References", "CR"]:
                    if alt in self.df.columns:
                        refs_col = alt
                        break
        
        if sep is None:
            sep = getattr(self, 'default_separator', '; ')
        
        # Compute disruption
        self.disruption_df = compute_document_disruption(
            self.df,
            id_col=id_col,
            refs_col=refs_col,
            sep=sep,
            verbose=verbose,
        )
        
        # Add to main dataframe
        if add_to_df:
            self.df = add_disruption_to_df(
                self.df, self.disruption_df, id_col,
                columns=['cd_index', 'di_index', 'interpretation']
            )
        
        # Add to entity count dataframes
        if add_to_counts:
            self._add_disruption_to_counts(verbose=verbose)
        
        # Save if res_folder is set
        if hasattr(self, 'res_folder') and self.res_folder:
            try:
                from biblium import utilsbib
                utilsbib.to_excel_fancy(
                    self.disruption_df,
                    f_name=os.path.join(self.res_folder, "tables", "disruption_documents.xlsx"),
                )
            except Exception as e:
                if verbose:
                    print(f"  Warning: Could not save disruption table: {e}")
        
        return self.disruption_df
    
    def _add_disruption_to_counts(self, verbose: bool = True):
        """Add disruption metrics to existing count dataframes."""
        if self.disruption_df is None:
            return
        
        id_col = self.mapping.get("unique-id", "unique-id")
        sep = getattr(self, 'default_separator', '; ')
        
        # Sources
        if hasattr(self, 'sources_counts_df') and self.sources_counts_df is not None:
            source_col = self.mapping.get("Source_title", "Source title")
            try:
                source_di = aggregate_disruption_by_entity(
                    self.disruption_df, self.df, source_col, id_col, sep="|||", min_docs=1
                )
                # Merge with existing counts - match on source name
                src_name_col = 'Source' if 'Source' in self.sources_counts_df.columns else source_col
                di_cols = ['cd_index_mean', 'cd_index_median', 'cd_index_std', 'di_index_mean']
                di_cols = [c for c in di_cols if c in source_di.columns]
                if di_cols:
                    source_di = source_di.rename(columns={source_col: src_name_col})
                    self.sources_counts_df = self.sources_counts_df.merge(
                        source_di[[src_name_col] + di_cols],
                        on=src_name_col, how='left'
                    )
                    if verbose:
                        print(f"  Added disruption to sources_counts_df")
            except Exception as e:
                if verbose:
                    print(f"  Warning: Could not add disruption to sources: {e}")
        
        # Authors
        if hasattr(self, 'authors_counts_df') and self.authors_counts_df is not None:
            author_col = self.mapping.get("Authors", "Authors")
            try:
                author_di = aggregate_disruption_by_entity(
                    self.disruption_df, self.df, author_col, id_col, sep=sep, min_docs=1
                )
                # Match on author name - try different column names
                auth_name_col = None
                for col_name in ['Author', 'Authors', author_col]:
                    if col_name in self.authors_counts_df.columns:
                        auth_name_col = col_name
                        break
                
                if auth_name_col is None:
                    if verbose:
                        print(f"  Warning: Could not find author name column in authors_counts_df")
                else:
                    di_cols = ['cd_index_mean', 'cd_index_median', 'cd_index_std', 'di_index_mean']
                    di_cols = [c for c in di_cols if c in author_di.columns]
                    if di_cols:
                        author_di = author_di.rename(columns={author_col: auth_name_col})
                        self.authors_counts_df = self.authors_counts_df.merge(
                            author_di[[auth_name_col] + di_cols],
                            on=auth_name_col, how='left'
                        )
                        if verbose:
                            print(f"  Added disruption to authors_counts_df")
            except Exception as e:
                if verbose:
                    print(f"  Warning: Could not add disruption to authors: {e}")
        
        # Countries
        if hasattr(self, 'all_countries_counts_df') and self.all_countries_counts_df is not None:
            # Try to find country column
            country_col = None
            for col in ["Countries", "Country", "All Countries", "Countries of Authors"]:
                if col in self.df.columns:
                    country_col = col
                    break
            
            if country_col:
                try:
                    country_di = aggregate_disruption_by_entity(
                        self.disruption_df, self.df, country_col, id_col, sep=sep, min_docs=1
                    )
                    # Match on country name
                    cnt_name_col = 'Country' if 'Country' in self.all_countries_counts_df.columns else country_col
                    di_cols = ['cd_index_mean', 'cd_index_median', 'cd_index_std', 'di_index_mean']
                    di_cols = [c for c in di_cols if c in country_di.columns]
                    if di_cols:
                        country_di = country_di.rename(columns={country_col: cnt_name_col})
                        self.all_countries_counts_df = self.all_countries_counts_df.merge(
                            country_di[[cnt_name_col] + di_cols],
                            on=cnt_name_col, how='left'
                        )
                        if verbose:
                            print(f"  Added disruption to all_countries_counts_df")
                except Exception as e:
                    if verbose:
                        print(f"  Warning: Could not add disruption to countries: {e}")
    
    def get_disruption_by_year(self) -> pd.DataFrame:
        """
        Get disruption index aggregated by year.
        
        Returns
        -------
        pd.DataFrame
            Year-level disruption statistics.
        """
        if self.disruption_df is None:
            self.compute_disruption_index()
        
        year_col = self.mapping.get("Year", "Year")
        id_col = self.mapping.get("unique-id", "unique-id")
        
        return aggregate_disruption_by_entity(
            self.disruption_df, self.df, year_col, id_col,
            sep="|||", min_docs=1
        )
    
    def get_top_disruptive(self, n: int = 20) -> pd.DataFrame:
        """
        Get top N most disruptive papers.
        
        Parameters
        ----------
        n : int
            Number of papers to return.
        
        Returns
        -------
        pd.DataFrame
            Top disruptive papers with metadata.
        """
        if self.disruption_df is None:
            self.compute_disruption_index()
        
        # Get top disruptive
        top = self.disruption_df.nlargest(n, 'cd_index')
        
        # Merge with main df for metadata
        id_col = self.mapping.get("unique-id", "unique-id")
        title_col = self.mapping.get("Title", "Title")
        year_col = self.mapping.get("Year", "Year")
        cit_col = self.mapping.get("Cited_by", "Cited by")
        
        self.df['_doc_id_join'] = self.df[id_col].astype(str).str.strip().str.lower()
        top['_doc_id_join'] = top['doc_id'].astype(str).str.strip().str.lower()
        
        cols_to_get = [c for c in [id_col, title_col, year_col, cit_col, '_doc_id_join'] 
                       if c in self.df.columns]
        
        result = top.merge(
            self.df[cols_to_get].drop_duplicates('_doc_id_join'),
            on='_doc_id_join', how='left'
        ).drop(columns=['_doc_id_join'])
        
        self.df = self.df.drop(columns=['_doc_id_join'])
        
        return result
    
    def get_top_consolidating(self, n: int = 20) -> pd.DataFrame:
        """
        Get top N most consolidating papers.
        
        Parameters
        ----------
        n : int
            Number of papers to return.
        
        Returns
        -------
        pd.DataFrame
            Top consolidating papers with metadata.
        """
        if self.disruption_df is None:
            self.compute_disruption_index()
        
        # Get top consolidating (lowest cd_index)
        top = self.disruption_df.nsmallest(n, 'cd_index')
        
        # Merge with main df for metadata
        id_col = self.mapping.get("unique-id", "unique-id")
        title_col = self.mapping.get("Title", "Title")
        year_col = self.mapping.get("Year", "Year")
        cit_col = self.mapping.get("Cited_by", "Cited by")
        
        self.df['_doc_id_join'] = self.df[id_col].astype(str).str.strip().str.lower()
        top['_doc_id_join'] = top['doc_id'].astype(str).str.strip().str.lower()
        
        cols_to_get = [c for c in [id_col, title_col, year_col, cit_col, '_doc_id_join'] 
                       if c in self.df.columns]
        
        result = top.merge(
            self.df[cols_to_get].drop_duplicates('_doc_id_join'),
            on='_doc_id_join', how='left'
        ).drop(columns=['_doc_id_join'])
        
        self.df = self.df.drop(columns=['_doc_id_join'])
        
        return result
    
    def get_source_disruption(self, min_docs: int = 3) -> pd.DataFrame:
        """
        Get average Disruption Index by source/journal.
        
        Parameters
        ----------
        min_docs : int
            Minimum documents for source to be included.
        
        Returns
        -------
        pd.DataFrame
            Source-level disruption statistics.
        """
        if self.disruption_df is None:
            self.compute_disruption_index()
        
        source_col = self.mapping.get("Source_title", "Source title")
        id_col = self.mapping.get("unique-id", "unique-id")
        
        return aggregate_disruption_by_entity(
            self.disruption_df, self.df,
            entity_col=source_col,
            id_col=id_col,
            sep="|||",  # Sources typically don't have separators
            min_docs=min_docs,
        )
    
    def get_author_disruption(self, min_docs: int = 2) -> pd.DataFrame:
        """
        Get average Disruption Index by author.
        
        Parameters
        ----------
        min_docs : int
            Minimum documents for author to be included.
        
        Returns
        -------
        pd.DataFrame
            Author-level disruption statistics.
        """
        if self.disruption_df is None:
            self.compute_disruption_index()
        
        author_col = self.mapping.get("Authors", "Authors")
        id_col = self.mapping.get("unique-id", "unique-id")
        sep = getattr(self, 'default_separator', '; ')
        
        return aggregate_disruption_by_entity(
            self.disruption_df, self.df,
            entity_col=author_col,
            id_col=id_col,
            sep=sep,
            min_docs=min_docs,
        )
    
    def get_country_disruption(self, min_docs: int = 5) -> pd.DataFrame:
        """
        Get average Disruption Index by country.
        
        Parameters
        ----------
        min_docs : int
            Minimum documents for country to be included.
        
        Returns
        -------
        pd.DataFrame
            Country-level disruption statistics.
        """
        if self.disruption_df is None:
            self.compute_disruption_index()
        
        # Try to find country column
        country_col = None
        for col in ["Countries", "Country", "All Countries", "CA Country"]:
            if col in self.df.columns:
                country_col = col
                break
        
        if country_col is None:
            country_col = self.mapping.get("Countries", "Countries")
        
        id_col = self.mapping.get("unique-id", "unique-id")
        sep = getattr(self, 'default_separator', '; ')
        
        return aggregate_disruption_by_entity(
            self.disruption_df, self.df,
            entity_col=country_col,
            id_col=id_col,
            sep=sep,
            min_docs=min_docs,
        )
    
    def get_affiliation_disruption(self, min_docs: int = 3) -> pd.DataFrame:
        """
        Get average Disruption Index by affiliation/institution.
        
        Parameters
        ----------
        min_docs : int
            Minimum documents for affiliation to be included.
        
        Returns
        -------
        pd.DataFrame
            Affiliation-level disruption statistics.
        """
        if self.disruption_df is None:
            self.compute_disruption_index()
        
        affiliation_col = self.mapping.get("Affiliations", "Affiliations")
        id_col = self.mapping.get("unique-id", "unique-id")
        sep = getattr(self, 'default_separator', '; ')
        
        return aggregate_disruption_by_entity(
            self.disruption_df, self.df,
            entity_col=affiliation_col,
            id_col=id_col,
            sep=sep,
            min_docs=min_docs,
        )
    
    def get_year_disruption(self) -> pd.DataFrame:
        """
        Get average Disruption Index by year.
        
        Returns
        -------
        pd.DataFrame
            Year-level disruption statistics.
        """
        if self.disruption_df is None:
            self.compute_disruption_index()
        
        year_col = self.mapping.get("Year", "Year")
        id_col = self.mapping.get("unique-id", "unique-id")
        
        return aggregate_disruption_by_entity(
            self.disruption_df, self.df,
            entity_col=year_col,
            id_col=id_col,
            sep="|||",  # Years don't have separators
            min_docs=1,
        )
